#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import sys

from paleomix.common.fileutils import open_ro
from paleomix.common.formats._common import FormatError
from paleomix.common.utilities import Immutable, TotallyOrdered


class FASTQError(FormatError):
    pass


class FASTQ(TotallyOrdered, Immutable):
    def __init__(self, name, meta, sequence, qualities):
        if not (name and isinstance(name, str)):
            raise FASTQError("FASTQ name must be a non-empty string")
        elif not (isinstance(meta, str) or (meta is None)):
            raise FASTQError("FASTQ meta must be a string, or None")
        elif not isinstance(sequence, str):
            raise FASTQError("FASTQ sequence must be a string")
        elif not isinstance(qualities, str):
            raise FASTQError("FASTQ qualities must be a string")
        elif len(sequence) != len(qualities):
            raise FASTQError("FASTQ sequence and qualities length differ")

        Immutable.__init__(
            self, name=name, meta=meta or "", sequence=sequence, qualities=qualities
        )

    def write(self, fileobj=sys.stdout):
        """Writes a FASTQ record to fileobj."""
        name = self.name
        if self.meta:
            name = "%s %s" % (name, self.meta)

        fileobj.write("@%s\n%s\n+\n%s\n" % (name, self.sequence, self.qualities))

    @classmethod
    def from_lines(cls, lines):
        """Parses FASTQ sequences found in a sequence of lines, and returns
        a tuple for each FASTQ record: ((name, meta-information), sequence)
        No assumptions are made about the line-lengths."""
        lines_iter = iter(lines)
        while True:
            try:
                header = next(lines_iter).rstrip()
                if not header:
                    break
            except StopIteration:
                break  # No sequences left

            if not header.startswith("@"):
                raise FASTQError("Invalid FASTQ header: %r" % (header,))

            header_fields = header.split(None, 1)
            name = header_fields[0]
            meta = header_fields[1] if len(header_fields) == 2 else None

            try:
                sequence = next(lines_iter).rstrip()
                separator = next(lines_iter).rstrip()
                qualities = next(lines_iter).rstrip()
            except StopIteration:
                raise FASTQError("Partial FASTQ record: %r" % (name,))

            if not separator.startswith("+"):
                raise FASTQError(
                    "Invalid FASTQ separator for %r; expected '+', found %r"
                    % (name, separator)
                )
            elif len(sequence) != len(qualities):
                raise FASTQError(
                    "Sequence length does not match qualities length for %r" % (name,)
                )

            yield FASTQ(
                name=name[1:], meta=meta, sequence=sequence, qualities=qualities
            )

    @classmethod
    def from_file(cls, filename):
        """Reads an unindexed FASTQ file, returning a sequence of
        tuples containing the name and sequence of each entry in
        the file. The FASTQ file may be GZIP/BZ2 compressed."""
        with open_ro(filename) as handle:
            yield from FASTQ.from_lines(handle)

    def __lt__(self, other):
        if not isinstance(other, FASTQ):
            return NotImplemented

        return (self.name, self.meta, self.sequence, self.qualities) < (
            other.name,
            other.meta,
            other.sequence,
            other.qualities,
        )

    def __hash__(self):
        return hash((self.name, self.meta, self.sequence, self.qualities))

    def __repr__(self):
        return "FASTQ(%r, %r, %r, %r)" % (
            self.name,
            self.meta,
            self.sequence,
            self.qualities,
        )


class FASTQualities:
    """Given a set of FASTQ records, this class attempts to identify the
    offset used to encode the quality scores pf those records.

    The following constants may be returned:
    - OFFSET_33: Offset identified as being 33
    - OFFSET_64: Offset identified as being 64
    - BOTH: Both offset 33 and 64 found, mixed file? (error)
    - MISSING: No quality scores found, wrong file? (error)
    - AMBIGUOUS: Qualities could be either offset. (warning)
    """

    # Quality score offsets for Phred (or similar) scores in FASTQ reads (33 or 64)
    OFFSET_33 = 33
    OFFSET_64 = 64
    # Quality score found in both ranges that are unique to each offset,
    # suggesting that the list contains mixed quality offsets, or invalid data.
    BOTH = "BOTH"
    # No quality scores in the expected range of ASCII values (33 .. 105)
    MISSING = "MISSING"
    # Quality scores are in the ASCII range 59 .. 74, which could signify
    # low-quality reads with offset 64, or high-quality reads with offset 33
    AMBIGUOUS = "AMBIGUOUS"

    def __init__(self):
        self._qualities = set()

    def update(self, record):
        self._qualities.update(record.qualities)

    def offsets(self):
        qualities = [False] * 256
        for quality in self._qualities:
            qualities[ord(quality)] = True

        # The range of scores that can unambiguously be identified
        # as belonging to Phred scores with offset 33 or 64. Scores
        # in between could potentially signify either offset
        # See e.g. http://en.wikipedia.org/wiki/FASTQ_format#Encoding
        has_offset_33_scores = any(qualities[33:59])
        has_ambiguous_scores = any(qualities[59:75])
        has_offset_64_scores = any(qualities[75:105])

        if has_offset_33_scores:
            if has_offset_64_scores:
                return FASTQualities.BOTH
            return FASTQualities.OFFSET_33
        elif has_offset_64_scores:
            return FASTQualities.OFFSET_64
        elif has_ambiguous_scores:
            return FASTQualities.AMBIGUOUS

        return FASTQualities.MISSING



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# Copyright (c) 2014 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import faulthandler
import logging
import signal
import sys

import pysam

import paleomix.common.logging
import paleomix.common.system

_COMMANDS = {
    # BAM/FASTQ pipeline
    "bam": "paleomix.pipelines.bam.pipeline",
    "bam_pipeline": "paleomix.pipelines.bam.pipeline",
    "trim": "paleomix.pipelines.bam.trim_pipeline",
    "trim_pipeline": "paleomix.pipelines.bam.trim_pipeline",
    # Phylogenetic pipeline
    "phylo": "paleomix.pipelines.phylo.pipeline",
    "phylo_pipeline": "paleomix.pipelines.phylo.pipeline",
    # Zonkey
    "zonkey": "paleomix.pipelines.zonkey.pipeline",
    "zonkey:db": "paleomix.pipelines.zonkey.build_db",
    "zonkey:mito": "paleomix.pipelines.zonkey.build_mito",
    "zonkey:tped": "paleomix.pipelines.zonkey.build_tped",
    # BAM file tools
    "cleanup": "paleomix.tools.cleanup",
    "coverage": "paleomix.tools.coverage",
    "depths": "paleomix.tools.depths",
    "dupcheck": "paleomix.tools.dupcheck",
    # VCF/etc. tools
    "rmdup_collapsed": "paleomix.tools.rmdup_collapsed",
    "vcf_filter": "paleomix.tools.vcf_filter",
    "vcf_to_fasta": "paleomix.tools.vcf_to_fasta",
    # Misc tools
    ":validate_fastq": "paleomix.tools.validate_fastq",
    ":rscript": "paleomix.tools.rscript",
}


_HELP = """PALEOMIX - pipelines and tools for NGS data analyses
Version: {version}

Pipelines:
    paleomix bam              -- Pipeline for trimming and mapping of NGS reads.
    paleomix trim             -- Equivalent to the 'bam' pipeline, but only runs
                                 the FASTQ trimming steps.
    paleomix phylo            -- Pipeline for genotyping and phylogenetic
                                 inference from BAMs.
    paleomix zonkey           -- Pipeline for detecting F1 (equine) hybrids.

BAM/SAM tools:
    paleomix coverage         -- Calculate coverage across reference sequences
                                 or regions of interest.
    paleomix depths           -- Calculate depth histograms across reference
                                 sequences or regions of interest.
    paleomix rmdup_collapsed  -- Filters PCR duplicates for collapsed paired-
                                 ended reads generated by the AdapterRemoval
                                 tool.

VCF/GTF/BED/Pileup tools:
    paleomix vcf_filter       -- Quality filters for VCF records, similar to
                                 'vcfutils.pl varFilter'.
    paleomix vcf_to_fasta     -- Create most likely FASTA sequence from tabix-
                                 indexed VCF file.

If you make use of PALEOMIX in your work, please cite
  Schubert et al, "Characterization of ancient and modern genomes by SNP
  detection and phylogenomic and metagenomic analysis using PALEOMIX".
  Nature Protocols. 2014 May; 9(5): 1056-82. doi: 10.1038/nprot.2014.063
"""


def main(argv):
    # Change process name from 'python' to 'paleomix'
    paleomix.common.system.set_procname("paleomix")
    # Setup basic logging to STDERR
    paleomix.common.logging.initialize_console_logging()
    # Silence log-messages from HTSLIB
    pysam.set_verbosity(0)

    # Add handlers to print backtraces on SIGSEGV, SIGFPE, SIGABRT, SIGBUS and SIGILL
    faulthandler.enable()
    # Add handler to print backtraces on SIGUSR1
    faulthandler.register(signal.SIGUSR1)

    if not argv or argv[0] in ("-h", "--help", "help"):
        print(_HELP.format(version=paleomix.__version__))
        return 0
    elif argv[0] in ("--version",):
        print("paleomix v{}".format(paleomix.__version__))
        return 0

    command = _COMMANDS.get(argv[0])
    if command is None:
        log = logging.getLogger(__name__)
        log.error("Unknown command %r", argv[0])
        return 1

    module = __import__(command, fromlist=["main"])

    return module.main(argv[1:])


def entry_point():
    return main(sys.argv[1:])


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



# Copyright (c) 2023 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import os
import re
import subprocess
import sys

import paleomix.common.argparse as argparse
from paleomix.common import resources


def parse_args(argv):
    parser = argparse.ArgumentParser(prog="paleomix :rscript")
    parser.add_argument("script", help="Name of Rscript bundled with PALEOMIX")
    parser.add_argument("args", nargs="*")

    return parser.parse_args(argv)


def main(argv):
    args = parse_args(argv)

    if not re.match(r"^[a-z]+/[a-z]+\.r$", args.script, flags=re.I):
        print(f"ERROR: Invalid resource {args.script!r}")
        return 1

    with resources.access(os.path.join("rscripts", args.script)) as path:
        result = subprocess.run(
            ("Rscript", path, *args.args),
            stdin=subprocess.DEVNULL,
            check=False,
        )

    return result.returncode


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/env python3
import configargparse

import paleomix

SUPPRESS = configargparse.SUPPRESS


class ArgumentDefaultsHelpFormatter(configargparse.ArgumentDefaultsHelpFormatter):
    """Modified ArgumentDefaultsHelpFormatter that excludes several constants (True,
    False, None) and uses a custom presentation of the default value.
    """

    def __init__(self, *args, **kwargs):
        # Enable wordwrapping
        kwargs.setdefault("width", 79)

        super().__init__(*args, **kwargs)

    def _get_help_string(self, action):
        # The following values look silly as part of a help string
        if isinstance(action.default, bool) or action.default in [None, [], ()]:
            return action.help

        # The subclass does not allow modification to the defaults string, so instead
        # we access the logic by simply checking if the result was modified.
        if super()._get_help_string(action) == action.help:
            return action.help

        return action.help + " [%(default)s]"


class ArgumentParser(configargparse.ArgumentParser):
    """Supports keys with underscores instead of dashes, for backwards compatibility
    with old paleomix config files, provided that these do not use per-host sections.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("formatter_class", ArgumentDefaultsHelpFormatter)

        super().__init__(*args, **kwargs)

        self.add_argument(
            "-v",
            "--version",
            action="version",
            version="%(prog)s v" + paleomix.__version__,
        )

    def get_possible_config_keys(self, *args, **kwargs):
        keys = super().get_possible_config_keys(*args, **kwargs)
        for key in keys:
            key = key.strip("-").replace("-", "_")
            if key not in keys:
                keys.append(key)

        return keys

    def convert_item_to_command_line_arg(self, action, key, value):
        # Ignore empty options from old config files
        if action and value in ("", "="):
            return []

        return super().convert_item_to_command_line_arg(action, key, value)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# Copyright (c) 2014 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Factory for AtomicCmdBuilders for the various PALEOMIX commands.

Ensures that the version called corresponds to the running version, in case
multiple versions are present in the users' PATH, or that the current version
is not available from the users' PATH.
"""

import sys

import paleomix.main
from paleomix.atomiccmd.builder import AtomicCmdBuilder
from paleomix.common import versions
from paleomix.common.utilities import safe_coerce_to_tuple

RSCRIPT_VERSION = versions.Requirement(
    call=("Rscript", "--version"),
    search=r"version (\d+)\.(\d+).(\d+)",
    checks=versions.GE(3, 3, 3),
    priority=10,
)


def command(args):
    interpreter = sys.executable
    script = paleomix.main.__file__

    return (interpreter, script, *safe_coerce_to_tuple(args))


def new(*args, **kwargs):
    """Returns AtomicCmdBuilder setup to call the tools accessible through the
    'paleomix' command-line tool. This builder adds executable / version checks
    for the specified command, but does not add any arguments. Thus, calling
    new with the argument "cat" produces the equivalent of ["paleomix", "cat"].
    """
    script = paleomix.main.__file__

    return AtomicCmdBuilder(command(args), AUX_PALEOMIX=script, **kwargs)


def rscript_command(args):
    return command((":rscript", *safe_coerce_to_tuple(args)))


def rscript(args, **kwargs):
    return new(
        ":rscript",
        *args,
        CHECK_R=RSCRIPT_VERSION,
        **kwargs,
    ).finalize()



#!/usr/bin/python3
#
# Copyright (c) 2016 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import copy
import os

import pysam

import paleomix
import paleomix.common.fileutils as fileutils
import paleomix.pipelines.zonkey.parts.admixture as admixture
import paleomix.pipelines.zonkey.parts.nuclear as nuclear
from paleomix.common import resources
from paleomix.node import Node
from paleomix.nodes.raxml import RAXML_VERSION
from paleomix.nodes.samtools import SAMTOOLS_VERSION
from paleomix.pipelines.zonkey.common import read_summary
from paleomix.tools.factory import RSCRIPT_VERSION


class ReportNode(Node):
    def __init__(self, config, root, has_nuc, has_mt, dependencies=()):
        """

        Arguments:
          config -- Config object generated using paleomix.pipelines.zonkey.config.
          root -- Root folder containing current analysis.
          has_nuc -- True if a nuclear BAM was provided.
          has_mt -- True if a mitochondrial BAM was provided.
          dependencies -- Nodes for ReportNode to depend on.
        """

        self._root = root
        self._data = copy.deepcopy(config.database)
        self._report = AnalysisReport(config, root, has_nuc, has_mt)
        self._has_nuc = bool(has_nuc)
        self._has_mt = bool(has_mt)

        self._treemix_outgroup = config.treemix_outgroup
        self._treemix_k = config.treemix_k
        if self._treemix_k is None:
            self._treemix_k = "<automatic>"

        Node.__init__(
            self,
            description="writing report to %s"
            % (os.path.join(self._root, "report.html"),),
            input_files=self._report.input_files(),
            output_files=(
                os.path.join(self._root, "report.html"),
                os.path.join(self._root, "report.css"),
            ),
            dependencies=dependencies,
        )

    def _run(self, _config, temp):
        with open(os.path.join(temp, "report.html"), "w") as output_handle:
            revision = self._data.settings["Revision"]
            header = _HTML_HEADER.format(
                Version=paleomix.__version__,
                Database=revision,
                Sidebar=self._build_sidebar(),
            )
            output_handle.write(header)

            self._write_intro_and_overview(output_handle)
            self._write_sample_description(output_handle)

            if self._has_nuc:
                self._write_admixture_estimates(output_handle)
                self._write_pca_plots(output_handle)
                self._write_treemix_plots(output_handle)

            if self._has_mt:
                self._write_mitochondrial_plots(output_handle)

            self._write_references(output_handle)

            output_handle.write(_HTML_FOOTER)

    def _teardown(self, config, temp):
        fileutils.make_dirs(self._root)

        fileutils.move_file(
            os.path.join(temp, "report.html"), os.path.join(self._root, "report.html")
        )

        resources.copy_resource(
            os.path.join("reports", "zonkey", "report.css"),
            os.path.join(self._root, "report.css"),
        )

    def _write_intro_and_overview(self, output_handle):
        output_handle.write(_SECTION_HEADER.format(name="intro", title="Introduction"))
        output_handle.write(_INTRODUCTION)

        output_handle.write(
            _SECTION_HEADER.format(name="overview", title="Analysis overview")
        )

        revision = self._data.settings["Revision"]
        overview = _OVERVIEW_HEADER.format(
            DATABASE=revision,
            PYSAM=pysam.__version__,
            SAMTOOLS=_fmt_v(SAMTOOLS_VERSION),
            PLINK=_fmt_v(nuclear.PLINK_VERSION),
            RSCRIPT=_fmt_v(RSCRIPT_VERSION),
        )
        output_handle.write(overview)

        if self._has_nuc:
            summary = self._report.snp_summary()
            output_handle.write(_OVERVIEW_NUCLEAR % summary)

        if self._has_mt:
            summary = self._report.mito_summary()
            output_handle.write(_OVERVIEW_MITOCHONDRIA % summary)

        if self._has_nuc:
            output_handle.write(_OVERVIEW_NUCLEAR_COVERAGE)

        output_handle.write(_OVERVIEW_FOOTER)

    def _write_sample_description(self, output_handle):
        output_handle.write(
            _SECTION_HEADER.format(name="samples", title="Reference Panel")
        )

        group_headers = []
        for k_value in sorted(self._data.groups):
            group_headers.append("          <th>Group(%i)</th>" % k_value)
        output_handle.write(_SAMPLE_LIST_HEADER % ("\n".join(group_headers),))

        for name, row in sorted(self._data.samples.items()):
            row = dict(row)
            groups = []
            for key in self._data.groups:
                groups.append("          <td>%s</td>" % self._data.groups[key][name])
            row["Groups"] = "\n".join(groups)

            pub = row["Publication"]
            if pub.startswith("http"):
                row["Publication"] = '<a href="{0}">Link</a>'.format(pub.strip())
            elif row["Publication"].startswith("doi:"):
                pub = pub[4:].strip()
                url = "https://doi.org/{}".format(pub)
                row["Publication"] = 'doi:<a href="{0}">{1}</a>'.format(url, pub)

            output_handle.write(_SAMPLE_LIST_ROW.format(**row))

        output_handle.write("      </table>\n")

    def _write_admixture_estimates(self, output_handle):
        header = _SECTION_HEADER.format(name="admixture", title="Admixture Estimates")
        output_handle.write(header)

        admixture_v = _fmt_v(nuclear.ADMIXTURE_VERSION)
        overview = _ADMIXTURE_OVERVIEW.format(ADMIXTURE=admixture_v)
        output_handle.write(overview)

        for k_groups in sorted(self._data.groups):
            summary_incl = self._build_admixture_cell(k_groups, True)
            summary_excl = self._build_admixture_cell(k_groups, False)

            output_handle.write(
                _ADMIXTURE_ROW.format(
                    K=k_groups, Incl_TS=summary_incl, Excl_TS=summary_excl
                )
            )

    def _build_admixture_cell(self, k_groups, incl_ts, cutoff=admixture.CUTOFF):
        try:
            groups = self._report.admixture_results(k_groups, incl_ts)
        except admixture.AdmixtureError as error:
            return _warn("ERROR: {}</strong".format(error))

        n_admixture_candidates = sum((value >= cutoff) for _, value in groups)
        if n_admixture_candidates < 2:
            return "<strong>No admixture detected.</strong>"

        lines = ["<strong>Possible hybridization detected:</strong>", "<ul>"]

        for group, value in groups:
            if value >= cutoff:
                name = " / ".join(sorted(group))

                lines.append("  <li>%s (%.2f%%)</li>" % (name, value * 100))

        lines.append("</ul>")

        if n_admixture_candidates != 2:
            lines.append(
                _warn(
                    "WARNING: %s-way admixture detected; this may "
                    "indicate a false-positive result!" % (n_admixture_candidates,)
                )
            )
            return "\n            ".join(lines)

        percentiles = self._report.admixture_percentiles(
            data=self._data, k_groups=k_groups, incl_ts_k=incl_ts
        )

        if not ("Lower" in percentiles or "Upper" in percentiles):
            lines.append(_warn("WARNING: Could not determine percentiles."))

            return "\n            ".join(lines)

        if "Lower" not in percentiles:
            percentiles["Lower"] = percentiles["Upper"]
            finale = (
                "; note that this is more simulated reads than what was "
                "processed in this analyses, potentially resulting in an "
                "overrestimating of percentages."
            )
        elif "Upper" not in percentiles:
            percentiles["Upper"] = percentiles["Lower"]
            finale = (
                "; note that this is fewer simulated reads than what was "
                "processed in this analyses, potentially resulting in an "
                "underrestimating of percentages"
            )
        else:
            finale = "."

        lower_pct = "%.1f" % (
            (1.0 - max(percentiles["Lower"]["Upper"], percentiles["Upper"]["Lower"]))
            * 100.0,
        )
        upper_pct = "%.1f" % (
            (1.0 - min(percentiles["Lower"]["Upper"], percentiles["Upper"]["Lower"]))
            * 100.0,
        )

        pct_range = lower_pct
        if lower_pct != upper_pct:
            pct_range = "%s - %s" % (lower_pct, upper_pct)

        lower_reads = min(
            percentiles["Lower"]["NReads"], percentiles["Upper"]["NReads"]
        )
        upper_reads = max(
            percentiles["Lower"]["NReads"], percentiles["Upper"]["NReads"]
        )

        reads_range = lower_reads
        if lower_reads != upper_reads:
            reads_range = "%s to %s" % (lower_reads, upper_reads)

        lines.append(
            "Admixture results fall within %s percent of those "
            "observed for simulated F1 %s / %s hybrids, based on "
            "%s randomly selected reads%s"
            % (
                pct_range,
                percentiles["Sample1"],
                percentiles["Sample2"],
                reads_range,
                finale,
            )
        )

        return "\n            ".join(lines)

    def _write_pca_plots(self, output_handle):
        output_handle.write(_SECTION_HEADER.format(name="pca", title="PCA Plots"))

        smartpca_v = _fmt_v(nuclear.SMARTPCA_VERSION)
        output_handle.write(_PCA_SECTION.format(SMARTPCA=smartpca_v))

    def _write_treemix_plots(self, output_handle):
        output_handle.write(
            _SECTION_HEADER.format(name="treemix", title="Treemix Plots")
        )

        outgroups = ""
        if self._treemix_outgroup:
            outgroups = ", ".join(map(repr, self._treemix_outgroup))
            outgroups = (
                ". The tree was rooted on the clade containing the "
                "sample(s) %s" % (outgroups)
            )

        treemix_v = _fmt_v(nuclear.TREEMIX_VERSION)
        overview = _TREEMIX_OVERVIEW.format(
            treemix_k=self._treemix_k, treemix_outgroup=outgroups, TREEMIX=treemix_v
        )
        output_handle.write(overview)

        for prefix in ("incl_ts", "excl_ts"):
            output_handle.write("<h2>%s</h2>\n" % (_TS_LABELS[prefix],))

            for n_edges in (0, 1):
                variance_file = os.path.join(
                    self._root, "figures", "treemix", "%s_%i_variance.txt"
                )

                with open(variance_file % (prefix, n_edges)) as handle:
                    variance = handle.read().strip()

                treemix_row = _TREEMIX_TREE_ROW.format(
                    Prefix=prefix, Edges=n_edges, Variance=variance
                )
                output_handle.write(treemix_row)

    def _write_mitochondrial_plots(self, output_handle):
        header = _SECTION_HEADER.format(
            name="mitochondria", title="Mitochondrial Phylogeny"
        )
        output_handle.write(header)
        raxml_v = _fmt_v(RAXML_VERSION)
        output_handle.write(_MITOCONDRIAL_SECTION.format(RAXML=raxml_v))

    def _write_references(self, output_handle):
        header = _SECTION_HEADER.format(name="references", title="References")
        output_handle.write(header)
        output_handle.write(_REFERENCES)

    def _build_sidebar(self):
        lines = [_SIDEBAR_HEADER]

        if self._has_nuc:
            lines.append(_SIDEBAR_NUCLEAR)

        if self._has_mt:
            lines.append(_SIDEBAR_MITO)

        lines.append(_SIDEBAR_FOOTER)

        return "\n".join(lines)


class AnalysisReport:
    def __init__(self, config, root, has_nuc, has_mt):
        self._has_nuc = bool(has_nuc)
        self._has_mt = bool(has_mt)
        self._config = config
        self._root = root
        self._data = config.database

    def input_files(self):
        input_files = [self._config.database.filename]
        if self._has_nuc:
            # Summary file generated while building TPED files
            input_files.append(
                os.path.join(self._root, "results", "plink", "common.summary")
            )

            for postfix in ("incl_ts", "excl_ts"):
                admix_root = os.path.join(self._root, "results", "admixture")

                # Include files showing proproation of ancestral populations,
                # which are required to build admixture figures in the reports.
                for k_groups in sorted(self._data.groups):
                    input_files.append(
                        os.path.join(admix_root, "%s.%i.Q" % (postfix, k_groups))
                    )

                # Include files tabulating variance explained by models
                figures_path = os.path.join(self._root, "figures", "treemix")
                for n_edges in (0, 1):
                    variance_path = os.path.join(
                        figures_path, "%s_%i_variance.txt" % (postfix, n_edges)
                    )

                    input_files.append(variance_path)

        if self._has_mt:
            input_files.append(
                os.path.join(self._root, "results", "mitochondria", "sequences.summary")
            )

        return input_files

    def snp_summary(self):
        return read_summary(
            os.path.join(self._root, "results", "plink", "common.summary")
        )

    def mito_summary(self):
        return read_summary(
            os.path.join(self._root, "results", "mitochondria", "sequences.summary")
        )

    def admixture_results(self, k_groups, incl_ts, cutoff=admixture.CUTOFF):
        prefix = "incl_ts" if incl_ts else "excl_ts"
        filename = os.path.join(
            self._root, "results", "admixture", "%s.%i.Q" % (prefix, k_groups)
        )

        return admixture.read_admixture_results(
            filename=filename,
            data=self._config.database,
            k_groups=k_groups,
            cutoff=cutoff,
        )

    def admixture_percentiles(self, data, k_groups, incl_ts_k, cutoff=admixture.CUTOFF):
        try:
            results = self.admixture_results(k_groups, incl_ts_k, cutoff)
        except admixture.AdmixtureError:
            return

        groups = [group for group, value in results if value >= cutoff]
        if len(groups) != 2:
            return

        (sample1,) = groups[0]
        (sample2,) = groups[1]
        delta = abs(max(value for _, value in results) - 0.5)
        summary = read_summary(
            os.path.join(self._root, "results", "plink", "common.summary")
        )

        return admixture.get_percentiles(
            data=data,
            sample1=sample1,
            sample2=sample2,
            nreads=summary["n_reads_used"],
            k_groups=k_groups,
            has_ts=incl_ts_k,
            value=delta,
        )


def _fmt_v(requirement):
    return ".".join(map(str, requirement.version))


def _warn(text):
    return "<div class='warning'>%s</div>" % (text,)


###############################################################################

_TS_LABELS = {"incl_ts": "Including transitions", "excl_ts": "Excluding transitions"}


###############################################################################

_HTML_HEADER = """<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
<title>PALEOMIX Zonkey v{Version} - db rev. {Database}</title>
<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1" />
<link href="report.css" rel="stylesheet" type="text/css" />
</head>
<body>
<a name="top" id="top"></a>
<center>
  <div id="header">
    <h1>PALEOMIX Zonkey v{Version} - db rev. {Database}</h1>
    <h2>A pipeline for detection of F1 hybrids in equids.</h2>
  </div>
  <div id="content">
    <p class="introduction">
      Schubert M, Ermini L, Sarkissian CD, J&oacute;nsson H, Ginolhac A,
      Schaefer R, Martin MD, Fern&aacute;ndez R, Kircher M, McCue M,
      Willerslev E, and Orlando L. "<strong>Characterization of ancient and
      modern genomes by SNP detection and phylogenomic and metagenomic analysis
      using PALEOMIX</strong>". Nat Protoc. 2014 May;9(5):1056-82. doi:<a
      href="https://doi.org/10.1038/nprot.2014.063">10.1038/nprot.2014.063</a>.
      Epub 2014 Apr 10. PubMed PMID: <a
      href="http://www.ncbi.nlm.nih.gov/pubmed/24722405">24722405</a>.
    </p>
{Sidebar}
    <div id="mainbar">
"""  # noqa: E501

_SECTION_HEADER = """      <h1><a name="{name}" id="{name}"></a>{title}</h1>
"""

_INTRODUCTION = """
      <div>
        <div>
          The Zonkey Pipeline is a easy-to-use pipeline designed for the
          analyses of low-coverage, ancient DNA derived from historical
          equid samples, with the purpose of determining the species of
          the sample, as well as determining possible hybridization between
          horses, zebras, and asses. This is accomplished by comparing one
          or more samples aligned against the <em>Equus caballus</em> 2.0
          reference sequence with a reference panel of modern equids,
          including wild and domesticated equids.
        </div>
        <br/>
        <div>
          For more information, please refer to the
          <a href="http://paleomix.readthedocs.org/en/stable/zonkey_pipeline/index.html">
            the documentation for the Zonkey pipeline
          </a>
          or
          <a href="http://paleomix.readthedocs.org/en/stable/">
            the documentation for the PALEOMIX pipeline,
          </a>
          on which the Zonkey pipeline is based.
        </div>
        <br/>
"""  # noqa: E501

_OVERVIEW_HEADER = """
        <div>
          Zonkey run using database rev. {DATABASE}. Data processed using
          <a href="https://github.com/pysam-developers/pysam">pysam</a> v{PYSAM},
          <a href="https://samtools.github.io/">SAMTools</a> v{SAMTOOLS}
          [<em>Li <em>et al.</em> 2009</em>] and
          <a href="http://pngu.mgh.harvard.edu/purcell/plink/">PLINK</a> v{PLINK}
          [<em>Purcell <em>et al.</em> 2007</em>]; plotting was carried out using
          <a href="https://www.r-project.org/">R</a> v{RSCRIPT}. Additional
          tools listed below.
        </div>
        <br/>
        <div style="display:table;width:100%;">
          <div style="display:table-cell;">
"""

_OVERVIEW_NUCLEAR = """
            <strong>Nuclear report from '<em>%(filename)s</em>'</strong>

            <table style="width:95%%">
              <tr>
                <td style="width:50%%;">Number of reads processed:</td>
                <td>%(n_reads)s</td>
              </tr>
              <tr>
                <td>Number of reads overlapping SNPs:</td>
                <td>%(n_reads_used)s</td>
              </tr>
              <tr>
                <td>Number of SNPs used (incl. transitions):</td>
                <td>%(n_sites_incl_ts)s</td>
              </tr>
              <tr>
                <td>Number of SNPs used (excl. transitions):</td>
                <td>%(n_sites_excl_ts)s</td>
              </tr>
            </table>
"""

_OVERVIEW_MITOCHONDRIA = """
            <br>

            <h4>Mitochondrial report from '<em>%(filename)s</em>'</h4>

            <table style="width:95%%">
              <tr>
                <td style="width:50%%;">Reference sequence used:</td>
                <td>%(sequence_name)s</td>
              </tr>
              <tr>
                <td>Reference sequence length:</td>
                <td>%(sequence_len)s</td>
              </tr>
              <tr>
                <td>Number of sites covered:</td>
                <td>%(covered_sites)s</td>
              </tr>
              <tr>
                <td>Percentage of sites covered:</td>
                <td>%(covered_pct)s</td>
              </tr>
              <tr>
                <td>Mean coverage per site:</td>
                <td>%(mean_coverage)s</td>
              </tr>
            </table>
"""


_OVERVIEW_NUCLEAR_COVERAGE = """
          </div>
          <div style="display:table-cell;width:25%;max-width:350px;">
           <div>
             <strong>Autosomes vs. sex-chromosomes:</strong>
           </div>
           <div>
             <a href="figures/coverage/coverage.pdf">
               <img src="figures/coverage/coverage.png"
                    style="vertical-align:top;horizontal-align:center;">
             </a>
           </div>
"""

_OVERVIEW_FOOTER = """
          </div>
      </div>
"""

_SAMPLE_LIST_HEADER = """
      <table summary="List of samples in the reference panel.">
        <tr>
          <th>ID</th>
%s
          <th>Species</th>
          <th>Sex</th>
          <th>Sample Name</th>
          <th>Publication</th>
        </tr>
"""

_SAMPLE_LIST_ROW = """
        <tr>
          <td>{ID}</td>
{Groups}
          <td><em>{Species}</em></td>
          <td>{Sex}</th>
          <td>{SampleID}</td>
          <td>{Publication}</td>
        </tr>
"""


_ADMIXTURE_OVERVIEW = """
      <p>
        Admixture proportions estimated using
        <a href="https://www.genetics.ucla.edu/software/admixture/">ADMIXTURE</a>
        v{ADMIXTURE} <em>[Alexander <em>et al.</em> 2009]</em>, using default
        parameters.
      </p>
"""


_ADMIXTURE_ROW = """
      <table summary="Admixture between sample and reference populations.">
        <tr>
          <th>{K} ancestral groups</th>
          <th>{K} ancestral groups, excluding transitions</th>
        </tr>
        <tr>
          <td>
            <a href="figures/admixture/incl_ts_k{K}.pdf">
                <img style="width:95%" src="figures/admixture/incl_ts_k{K}.png"
                     alt="Admixture for k={K}, including transitions." />
            </a>
            <br/>
            <br/>
            {Incl_TS}
          </td>
          <td>
            <a href="figures/admixture/excl_ts_k{K}.pdf">
                <img style="width:95%" src="figures/admixture/excl_ts_k{K}.png"
                     alt="Admixture for k={K}, excluding transitions." />
            </a>
            <br/>
            <br/>
            {Excl_TS}
          </td>
        </tr>
      </table>
"""


_PCA_SECTION = """
      <p>
        Principal Component Analysis carried out using SmartPCA v{SMARTPCA},
        from the <a
        href="http://www.hsph.harvard.edu/alkes-price/software/">EIGENSOFT</a>
        toolkit.
      </p>

      <table summary="PCA plots comparing sample with the reference panel.">
        <tr>
          <th>Including transitions</th>
          <th>Excluding transitions</th>
        </tr>
        <tr>
          <td>
            <a href="figures/pca/incl_ts.pdf">
                <img style="width:95%" src="figures/pca/incl_ts.png"
                     alt="PCA plot, including transitions." />
            </a>
          </td>
          <td>
            <a href="figures/pca/excl_ts.pdf">
                <img style="width:95%" src="figures/pca/excl_ts.png"
                     alt="PCA plot, excluding transitions." />
            </a>
          </td>
        </tr>
      </table>
"""


_TREEMIX_OVERVIEW = """
      <p>
        Detection of population mixture using
        <a href="https://bitbucket.org/nygcresearch/treemix/wiki/Home">TreeMix</a>
        v{TREEMIX} <em>[Pickrell and Pritchard 2012]</em>; parameters were -k
        {treemix_k}; -global; and supervised estimation using ancestral groups
        listed in the Reference Panel{treemix_outgroup}.
      </p>
"""


_TREEMIX_TREE_ROW = """
      <table summary="Treemix plots, for {Edges} edge(s).">
        <tr>
          <th>Edges = {Edges}</th>
          <th>Residuals</th>
        </tr>
        <tr>
          <td>
            <a href="figures/treemix/{Prefix}_{Edges}_tree.pdf">
                <img style="width:95%"
                     src="figures/treemix/{Prefix}_{Edges}_tree.png"
                     alt="Treemix plot, {Edges} edge(s), incl. transitions." />
            </a>
          </td>
          <td>
            <a href="figures/treemix/{Prefix}_{Edges}_residuals.pdf">
                <img style="width:95%"
                     src="figures/treemix/{Prefix}_{Edges}_residuals.png"
                     alt="Treemix plot, {Edges} edge(s), excl. transitions." />
            </a>
          </td>
        </tr>
      </table>

      <p>
        Variance explained by model = {Variance}.
      </p>
"""


_MITOCONDRIAL_SECTION = """
      <p>
        Phylogenetic inference performed using RAxML
        v{RAXML} [<em>Stamatakis 2006</em>].
      <p>

      <div style="text-align:center;">
        <a href="figures/mitochondria/mito_phylo.pdf">
          <img src="figures/mitochondria/mito_phylo.png"
               alt="Mitochondrial maximum likelihood phylogeny." />
        </a>
      </div>
"""


_REFERENCES = """
        <p>
          <ul>
            <li>
              Alexander <em>et al</em>. "<strong>Fast model-based estimation of
              ancestry in unrelated individuals</strong>". <em>Genome Res</em>.
              2009 Sep;19(9):1655-64.
                doi:<a href="https://doi.org/10.1101/gr.094052.109">
                    10.1101/gr.094052.109</a>.
                PMID:<a href="http://www.ncbi.nlm.nih.gov/pubmed/19648217">
                  19648217</a>.
            </li>
            <li>
              Li <em>et al</em>. "<strong>The Sequence Alignment/Map format and
              SAMtools</strong>. <em>Bioinformatics</em>. 2009 Aug
              15;25(16):2078-9.
                doi:<a href="https://doi.org/10.1093/bioinformatics/btp352">
                    10.1093/bioinformatics/btp352</a>.
                PMID:<a href="http://www.ncbi.nlm.nih.gov/pubmed/19505943">
                  19505943</a>.
            </li>
            <li>
              Pickrell and Pritchard. "<strong>Inference of population splits
              and mixtures from genome-wide allele frequency data</strong>".
              <em>PLoS Genet</em>. 2012;8(11):e1002967.
              doi:<a href="https://doi.org/10.1371/journal.pgen.1002967">
                  10.1371/journal.pgen.1002967</a>.
              PMID:<a href="http://www.ncbi.nlm.nih.gov/pubmed/23166502">
                23166502</a>.
            </li>
            <li>
              Purcell <em>et al</em>. "<strong>PLINK: a tool set for whole-
              genome association and population-based linkage
              analyses</strong>". <em>Am J Hum Genet</em>. 2007
              Sep;81(3):559-75.
              PMID:<a href="http://www.ncbi.nlm.nih.gov/pubmed/17701901">
                17701901</a>.
            </li>
            <li>
              Stamatakis. "<strong>RAxML-VI-HPC: maximum likelihood-based
              phylogenetic analyses with thousands of taxa and mixed
              models</strong>". <em>Bioinformatics</em>. 2006 Nov
              1;22(21):2688-90. Epub 2006 Aug 23.
              doi:<a href="https://doi.org/10.1093/bioinformatics/btl446">
                  10.1093/bioinformatics/btl446</a>.
              PMID:<a href="http://www.ncbi.nlm.nih.gov/pubmed/16928733">
                16928733</a>.
          </ul>
        </p>
"""


_HTML_FOOTER = """
    </div>
  </div>
  <div id="footer">
    This report is based on the PLAIN 1.0 design by
    <a href="http://www.sixshootermedia.com/">6ix Shooter Media</a>,
    Creative Commons license.<br/>
  </div>
</center>
</body>
</html>
"""

###############################################################################

_SIDEBAR_HEADER = """
    <div id="sidebar">
      <h1>Contents</h1>
      <div class="submenu">
        <a href="#">Top</a>
        <a href="#intro">Introduction</a>
        <a href="#overview">Analysis overview</a>
        <a href="#samples">Reference Panel</a>
"""

_SIDEBAR_NUCLEAR = """
        <a href="#admixture">Admixture Estimates</a>
        <a href="#pca">PCA Plots</a>
        <a href="#treemix">Treemix Analyses</a>
"""

_SIDEBAR_MITO = """
        <a href="#mito_phylo">MT Phylogeny</a>
"""

_SIDEBAR_FOOTER = """
        <a href="#references">References</a>
      </div>
    </div>
"""



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import bz2
import errno
import gzip
import os
import shutil
import stat
from pathlib import Path
from typing import Any
from unittest.mock import ANY, DEFAULT, Mock, call, patch

import pytest

from paleomix.common.fileutils import (
    add_postfix,
    copy_file,
    create_temp_dir,
    describe_files,
    describe_paired_files,
    fspath,
    make_dirs,
    missing_executables,
    missing_files,
    move_file,
    open_ro,
    reroot_path,
    swap_ext,
    try_remove,
    try_rmtree,
)
from paleomix.common.testing import SetWorkingDirectory

###############################################################################
###############################################################################
# Tests for 'add_postfix'


def test_add_postfix__no_postfix() -> None:
    assert add_postfix("name.foo", "") == "name.foo"


def test_add_postfix__dot_postfix() -> None:
    assert add_postfix("name.foo", ".pf") == "name.pf.foo"


def test_add_postfix__underscore_postfix() -> None:
    assert add_postfix("name.foo", "_pf") == "name_pf.foo"


def test_add_postfix__no_ext__no_postfix() -> None:
    assert add_postfix("name", "") == "name"


def test_add_postfix__no_ext__dot_postfix() -> None:
    assert add_postfix("name", ".pf") == "name.pf"


def test_add_postfix__no_ext__underscore_postfix() -> None:
    assert add_postfix("name", "_pf") == "name_pf"


###############################################################################
###############################################################################
# Tests for 'swap_ext'


def test_swap_ext__has_ext_vs_empty_ext() -> None:
    assert swap_ext("name.foo", "") == "name"


def test_swap_ext__empty_ext_vs_empty_ext() -> None:
    assert swap_ext("name", "") == "name"


def test_swap_ext__has_ext_vs_dot_ext() -> None:
    assert swap_ext("name.foo", ".") == "name"


def test_swap_ext__dot_ext_vs_dot_ext() -> None:
    assert swap_ext("name.", ".") == "name"


def test_swap_ext__multiple__has_ext_vs_empty_ext() -> None:
    assert swap_ext("name.foo.bar", "") == "name.foo"


def test_swap_ext__multiple__has_ext_vs_dot_ext() -> None:
    assert swap_ext("name.foo.bar", ".") == "name.foo"


def test_swap_ext__multiple__dot_ext_vs_dot_ext() -> None:
    assert swap_ext("name.foo.", ".") == "name.foo"


def test_swap_ext__has_ext_vs_new_ext() -> None:
    assert swap_ext("name.foo", "bar") == "name.bar"


def test_swap_ext__has_ext_vs_new_dot_ext() -> None:
    assert swap_ext("name.foo", ".bar") == "name.bar"


def test_swap_ext__empty_ext_vs_new_ext() -> None:
    assert swap_ext("name", "bar") == "name.bar"


def test_swap_ext__dot_ext_vs_new_dot_ext() -> None:
    assert swap_ext("name", ".bar") == "name.bar"


###############################################################################
###############################################################################
# Tests for 'reroot_path'


def test_reroot_path__empty_root() -> None:
    assert reroot_path("", "/etc/apt/sources.list") == "sources.list"


def test_reroot_path__empty_path() -> None:
    assert reroot_path("/etc/apt", "") == "/etc/apt/"


def test_reroot_path__abs_abs__wo_final_dash() -> None:
    assert reroot_path("/etc/apt", "/tmp/sources.list") == "/etc/apt/sources.list"


def test_reroot_path__abs_abs__w_final_dash() -> None:
    assert reroot_path("/etc/apt/", "/tmp/sources.list") == "/etc/apt/sources.list"


def test_reroot_path__abs_rel__wo_final_dash() -> None:
    assert reroot_path("/etc/apt", "tmp/sources.list") == "/etc/apt/sources.list"


def test_reroot_path__abs_rel__w_final_dash() -> None:
    assert reroot_path("/etc/apt/", "tmp/sources.list") == "/etc/apt/sources.list"


def test_reroot_path__rel_abs__wo_final_dash() -> None:
    assert reroot_path("etc/apt", "/tmp/sources.list") == "etc/apt/sources.list"


def test_reroot_path__rel_abs__w_final_dash() -> None:
    assert reroot_path("etc/apt/", "/tmp/sources.list") == "etc/apt/sources.list"


def test_reroot_path__rel_rel__wo_final_dash() -> None:
    assert reroot_path("etc/apt", "tmp/sources.list") == "etc/apt/sources.list"


def test_reroot_path__rel_rel__w_final_dash() -> None:
    assert reroot_path("etc/apt/", "tmp/sources.list") == "etc/apt/sources.list"


###############################################################################
###############################################################################
# Tests for 'create_temp_dir'


def test_create_temp_dir__create(tmp_path: Path) -> None:
    tmp_dir_1 = create_temp_dir(tmp_path)
    tmp_dir_2 = create_temp_dir(tmp_path)
    assert os.path.exists(tmp_dir_1)
    assert os.path.exists(tmp_dir_2)


def test_create_temp_dir__empty(tmp_path: Path) -> None:
    tmp_dir = create_temp_dir(tmp_path)
    contents = os.listdir(tmp_dir)
    assert not contents


def test_create_temp_dir__permissions(tmp_path: Path) -> None:
    tmp_dir = create_temp_dir(tmp_path)
    stats = os.stat(tmp_dir)
    assert stats.st_mode & 0 == 0


def test_create_temp_dir__creation_preempted(tmp_path: Path) -> None:
    makedirs = os.makedirs
    mock = Mock(wraps=makedirs)
    # Raise exception first, then pass second call to wrapped function
    mock.side_effect = [OSError(errno.EEXIST, "dir exists"), DEFAULT]
    with patch("os.makedirs", mock):
        work_dir = create_temp_dir(tmp_path)

    assert work_dir.startswith(str(tmp_path))

    # Must be called with different directories
    assert mock.mock_calls == [call(ANY, mode=0o750), call(ANY, mode=0o750)]
    call_1, call_2 = mock.mock_calls
    assert call_1[1] != call_2[1]


def test_create_temp_dir__permission_denied() -> None:
    with patch("os.makedirs") as mock:
        mock.side_effect = OSError(errno.EACCES, "Simulated permission denied")

        with pytest.raises(OSError, match="Simulated permission denied"):
            create_temp_dir("/tmp")


###############################################################################
###############################################################################
# Tests for 'missing_files'


def test_missing_files__file_exists(tmp_path) -> None:
    file_1 = tmp_path / "file_1"
    file_1.touch()

    assert missing_files([file_1]) == []


def test_missing_files__file_doesnt_exist(tmp_path) -> None:
    file_1 = tmp_path / "file_1"

    assert missing_files([file_1]) == [str(file_1)]


def test_missing_files__mixed_files(tmp_path) -> None:
    file_1 = tmp_path / "file_1"
    file_2 = tmp_path / "file_2"
    file_1.touch()

    assert missing_files([file_1, file_2]) == [str(file_2)]


###############################################################################
###############################################################################
# Tests for 'missing_executables'


def test_missing_executables__executable() -> None:
    assert missing_executables(["ls"]) == []


def test_missing_executables__non_executable() -> None:
    assert missing_executables(["lsxxxx"]) == ["lsxxxx"]


def test_missing_executables__mixed() -> None:
    assert missing_executables(["lsxxxx", "ls"]) == ["lsxxxx"]


def test_missing_executables__with_path__missing(tmp_path: Path) -> None:
    executable = tmp_path / "ls"
    assert missing_executables([executable]) == [str(executable)]


def test_missing_executables__with_path__exists(tmp_path: Path) -> None:
    executable = tmp_path / "ls"
    executable.touch()
    executable.chmod(0o700)

    assert missing_executables([executable]) == []


###############################################################################
###############################################################################
# Tests for 'make_dirs'


def test_make_dirs__create_dir(tmp_path: Path) -> None:
    assert not any(tmp_path.iterdir())
    assert make_dirs(tmp_path / "test123")
    assert os.listdir(str(tmp_path)) == ["test123"]


def test_make_dirs__return_values(tmp_path: Path) -> None:
    assert make_dirs(tmp_path / "test234")
    assert not make_dirs(tmp_path / "test234")


def test_make_dirs__subdirs_return_values(tmp_path: Path) -> None:
    assert make_dirs(tmp_path / "test")
    assert make_dirs(tmp_path / "test" / "234")
    assert not make_dirs(tmp_path / "test" / "234")


def test_make_dirs__sub_directories(tmp_path: Path) -> None:
    assert not any(tmp_path.iterdir())
    assert make_dirs(tmp_path / "test" / "123")
    assert os.listdir(str(tmp_path)) == ["test"]
    assert os.listdir(str(tmp_path / "test")) == ["123"]


def test_make_dirs__permissions(tmp_path: Path) -> None:
    work_dir = tmp_path / "test_1"
    assert make_dirs(work_dir, mode=0o511)
    stats = work_dir.stat()
    assert oct(stats.st_mode & 0o777) == oct(0o511)


def test_make_dirs__creation_preemted(tmp_path: Path) -> None:
    makedirs = os.makedirs

    def _wrap_os_makedirs(*args: Any, **kwargs: Any):
        # Simulate somebody else creating the directory first
        makedirs(*args, **kwargs)
        makedirs(*args, **kwargs)

    with patch("os.makedirs", _wrap_os_makedirs):
        work_folder = tmp_path / "test"
        assert not make_dirs(work_folder)
        assert work_folder.exists()
        assert os.listdir(str(tmp_path)) == ["test"]


@pytest.mark.skip(reason="fails when using `act`")
def test_make_dirs__permission_denied(tmp_path: Path) -> None:
    # Make temporary folder read-only
    mode = tmp_path.stat().st_mode
    ro_mode = mode & ~(stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH)
    tmp_path.chmod(ro_mode)
    # Non OEXIST errors should be re-raised:
    with pytest.raises(OSError):
        make_dirs(tmp_path / "foo")


def test_make_dirs__empty_directory() -> None:
    with pytest.raises(ValueError, match="Empty directory passed to make_dirs"):
        make_dirs("")


###############################################################################
###############################################################################
# Tests for 'move_file'


def test_move_file__simple_move(tmp_path: Path) -> None:
    file_1 = tmp_path / "file_1"
    file_2 = tmp_path / "file_2"
    assert os.listdir(str(tmp_path)) == []
    file_1.write_text("1")
    assert os.listdir(str(tmp_path)) == ["file_1"]
    move_file(file_1, file_2)
    assert os.listdir(str(tmp_path)) == ["file_2"]
    assert file_2.read_text() == "1"


def test_move_file__simple_move_in_cwd(tmp_path: Path) -> None:
    with SetWorkingDirectory(tmp_path):
        assert os.listdir(".") == []
        (tmp_path / "file_1").write_text("1")
        assert os.listdir(".") == ["file_1"]
        move_file("file_1", "file_2")
        assert os.listdir(".") == ["file_2"]
        assert (tmp_path / "file_2").read_text() == "1"


@pytest.mark.skip(reason="fails when using `act`")
def test_move_dirs__permission_denied(tmp_path: Path) -> None:
    dst_folder = tmp_path / "dst"
    file_1 = tmp_path / "file"
    file_2 = dst_folder / "file"
    file_1.write_text("1")

    # Make destination folder read-only
    assert make_dirs(tmp_path / "dst")
    mode = dst_folder.stat().st_mode
    ro_mode = mode & ~(stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH)
    dst_folder.chmod(ro_mode)

    # Non ENOENT errors should be re-raised:
    with pytest.raises(IOError):
        move_file(file_1, file_2)


def test_move_file__move_to_existing_folder(tmp_path: Path) -> None:
    assert make_dirs(tmp_path / "src")
    assert make_dirs(tmp_path / "dst")
    file_1 = tmp_path / "src" / "file_1"
    file_2 = tmp_path / "dst" / "file_2"
    file_1.write_text("2")
    move_file(file_1, file_2)

    assert os.listdir(str(file_1.parent)) == []
    assert os.listdir(str(file_2.parent)) == ["file_2"]
    assert file_2.read_text() == "2"


def test_move_file__move_to_new_folder(tmp_path: Path) -> None:
    assert make_dirs(tmp_path / "src")
    file_1 = tmp_path / "src" / "file_1"
    file_2 = tmp_path / "dst" / "file_2"
    file_1.write_text("2")

    move_file(file_1, file_2)

    assert os.listdir(str(file_1.parent)) == []
    assert os.listdir(str(file_2.parent)) == ["file_2"]
    assert file_2.read_text() == "2"


def test_move_file__move_to_different_folder(tmp_path: Path) -> None:
    (tmp_path / "file_1").write_text("3")

    with SetWorkingDirectory(tmp_path):
        move_file("file_1", "dst/file_1")

    assert os.listdir(str(tmp_path)), ["dst"]
    assert os.listdir(str(tmp_path / "dst")), ["file_1"]
    assert (tmp_path / "dst" / "file_1").read_text() == "3"


def test_move_file__overwrite(tmp_path: Path) -> None:
    (tmp_path / "file_1").write_text("4")
    (tmp_path / "file_2").write_text("5")

    with SetWorkingDirectory(tmp_path):
        move_file("file_1", "file_2")

    assert os.listdir(str(tmp_path)) == ["file_2"]
    assert (tmp_path / "file_2").read_text() == "4"


def test_move_file__enoent_reraised_if_not_due_to_missing_folder() -> None:
    with pytest.raises(IOError):
        move_file("", "./dst")


def test_move_file__destination_removed_if_out_of_space(tmp_path) -> None:
    _shutil_move = shutil.move

    def _move(source, destination):
        _shutil_move(source, destination)
        raise OSError(errno.ENOSPC, "Out of space")

    source = tmp_path / "source"
    source.write_text("...")
    destination = tmp_path / "destination"

    with pytest.raises(OSError, match="Out of space"):
        with patch("shutil.move", wraps=_move):
            move_file(source, destination)

    assert not source.exists()
    assert not destination.exists()


###############################################################################
###############################################################################
# Tests for 'copy_file'


def test_copy_file__simple_copy(tmp_path: Path) -> None:
    file_1 = tmp_path / "file_1"
    file_2 = tmp_path / "file_2"
    file_1.write_text("1")
    assert os.listdir(str(tmp_path)) == ["file_1"]
    copy_file(file_1, file_2)
    assert set(os.listdir(str(tmp_path))) == set(["file_1", "file_2"])
    assert file_1.read_text() == "1"
    assert file_2.read_text() == "1"


def test_copy_file__simple_copy_in_cwd(tmp_path: Path) -> None:
    file_1 = tmp_path / "file_1"
    file_2 = tmp_path / "file_2"
    file_1.write_text("1")

    with SetWorkingDirectory(tmp_path):
        copy_file("file_1", "file_2")

    assert set(os.listdir(str(tmp_path))) == set(["file_1", "file_2"])
    assert file_1.read_text() == "1"
    assert file_2.read_text() == "1"


def test_copy_file__copy_to_existing_folder(tmp_path: Path) -> None:
    assert make_dirs(tmp_path / "src")
    assert make_dirs(tmp_path / "dst")
    file_1 = tmp_path / "src" / "file_1"
    file_2 = tmp_path / "dst" / "file_2"
    file_1.write_text("2")
    copy_file(file_1, file_2)
    assert os.listdir(str(file_1.parent)) == ["file_1"]
    assert os.listdir(str(file_2.parent)) == ["file_2"]
    assert file_1.read_text() == "2"
    assert file_2.read_text() == "2"


def test_copy_file__copy_to_new_folder(tmp_path: Path) -> None:
    assert make_dirs(tmp_path / "src")
    file_1 = tmp_path / "src" / "file_1"
    file_2 = tmp_path / "dst" / "file_2"
    file_1.write_text("2")
    copy_file(file_1, file_2)
    assert os.listdir(str(file_1.parent)) == ["file_1"]
    assert os.listdir(str(file_2.parent)) == ["file_2"]
    assert file_1.read_text() == "2"
    assert file_2.read_text() == "2"


def test_copy_file__copy_to_different_folder(tmp_path: Path) -> None:
    (tmp_path / "file_1").write_text("3")

    with SetWorkingDirectory(tmp_path):
        copy_file("file_1", "dst/file_1")

    assert set(os.listdir(str(tmp_path))) == set(["file_1", "dst"])
    assert os.listdir(str(tmp_path / "dst")) == ["file_1"]
    assert (tmp_path / "file_1").read_text() == "3"
    assert (tmp_path / "dst" / "file_1").read_text() == "3"


def test_copy_file__overwrite(tmp_path: Path) -> None:
    file_1 = tmp_path / "file_1"
    file_1.write_text("4")
    file_2 = tmp_path / "file_2"
    file_2.write_text("5")

    with SetWorkingDirectory(tmp_path):
        copy_file("file_1", "file_2")

    assert set(os.listdir(str(tmp_path))) == set(["file_1", "file_2"])
    assert file_1.read_text() == "4"
    assert file_2.read_text() == "4"


def test_copy_file__enoent_reraised_if_not_due_to_missing_folder() -> None:
    with pytest.raises(IOError):
        copy_file("", "./dst")


def test_copy_file__destination_removed_if_out_of_space(tmp_path) -> None:
    _shutil_copy = shutil.copy

    def _copy(source, destination):
        _shutil_copy(source, destination)
        raise OSError(errno.ENOSPC, "Out of space")

    source = tmp_path / "source"
    source.write_text("...")
    destination = tmp_path / "destination"

    with pytest.raises(OSError, match="Out of space"):
        with patch("shutil.copy", wraps=_copy):
            copy_file(source, destination)

    assert source.exists()
    assert not destination.exists()


###############################################################################
###############################################################################
# Tests for 'open'

_FASTA_TEXT = ">This_is_FASTA!\nACGTN\n>This_is_ALSO_FASTA!\nCGTNA\n"
_FASTA_BYTES = _FASTA_TEXT.encode("utf-8")


@pytest.mark.parametrize("func", (open, gzip.open, bz2.open))
def test_open_ro(func, tmp_path) -> None:
    filename = tmp_path / "file.fasta"
    with func(fspath(filename), "wt") as handle:
        handle.write(_FASTA_TEXT)

    with open_ro(filename) as handle:
        assert handle.read() == _FASTA_TEXT


@pytest.mark.parametrize("mode", ("r", "rt"))
@pytest.mark.parametrize("func", (open, gzip.open, bz2.open))
def test_open_ro__mode(func, mode, tmp_path) -> None:
    filename = tmp_path / "file.fasta"
    with func(fspath(filename), "wt") as handle:
        handle.write(_FASTA_TEXT)

    with open_ro(filename, mode) as handle:
        assert handle.read() == _FASTA_TEXT


def test_open_ro__invalid_mode() -> None:
    with pytest.raises(ValueError, match="foo"):
        open_ro("file.fasta", "foo")


@pytest.mark.parametrize("func", (open, gzip.open, bz2.open))
def test_open_ro__binary(func, tmp_path) -> None:
    filename = tmp_path / "file.fasta"
    with func(fspath(filename), "wt") as handle:
        handle.write(_FASTA_TEXT)

    with open_ro(filename, "rb") as handle:
        assert handle.read() == _FASTA_BYTES


class OddException(RuntimeError):
    pass


def test_open_ro__close_handle_on_error() -> None:
    mocks = Mock()
    mocks.file.read.side_effect = OddException("ARGH!")
    mocks.file.__enter__ = Mock(return_value=mocks.file)
    mocks.file.__exit__ = Mock(return_value=None)
    mocks.open.return_value = mocks.file

    with patch("builtins.open", mocks.open):
        with pytest.raises(OddException):
            open_ro("/var/abc")

    mocks.assert_has_calls(
        [
            call.open("/var/abc", "rb"),
            call.file.__enter__(),
            call.file.read(2),
            call.file.__exit__(ANY, ANY, ANY),
        ]
    )


###############################################################################
###############################################################################
# Tests for 'try_remove'


def test_try_remove(tmp_path: Path) -> None:
    fpath = tmp_path / "test.txt"
    fpath.write_text("1 2 3")
    assert try_remove(fpath)
    assert not fpath.exists()


def test_try_remove__missing(tmp_path: Path) -> None:
    fpath = tmp_path / "test.txt"
    assert not try_remove(fpath)
    assert not fpath.exists()


@pytest.mark.skip(reason="fails on OSX; throws PermissionError")
def test_try_remove__non_file(tmp_path: Path) -> None:
    with pytest.raises(OSError):
        try_remove(tmp_path)


###############################################################################
###############################################################################
# Tests for 'try_rmtree'


def test_try_rmtree(tmp_path: Path) -> None:
    fpath = tmp_path / "testdir"
    fpath.mkdir()
    (fpath / "file").write_text("1 2 3")
    assert try_rmtree(fpath)
    assert not fpath.exists()


def test_try_treedir__missing(tmp_path: Path) -> None:
    fpath = tmp_path / "testdir"
    assert not try_rmtree(fpath)
    assert not fpath.exists()


###############################################################################
###############################################################################
# Tests for 'describe_files'


def test_describe_files__no_files() -> None:
    assert describe_files(()) == "No files"


def test_describe_files__single_file() -> None:
    fpath = "/var/foo/bar"
    assert describe_files((fpath,)) == repr(fpath)


def test_describe_files__same_path_abs__3_differences() -> None:
    fpaths = ("/var/foo/bar", "/var/foo/foo")
    assert describe_files(fpaths) == "2 files in '/var/foo'"


def test_describe_files__same_path_abs__2_differences() -> None:
    fpaths = ("/var/foo/faz", "/var/foo/foo")
    assert describe_files(fpaths) == "'/var/foo/f??'"


def test_describe_files__same_path_abs__1_differences() -> None:
    fpaths = ("/var/foo/faz", "/var/foo/fao")
    assert describe_files(fpaths) == "'/var/foo/fa?'"


def test_describe_files__different_paths_abs() -> None:
    fpaths = ("/var/foo/bar", "/var/bar/foo")
    assert describe_files(fpaths) == "2 files"


def test_describe_files__same_path_rel() -> None:
    fpaths = ("var/foo/bar", "var/foo/foo")
    assert describe_files(fpaths) == "2 files in 'var/foo'"


def test_describe_files__different_paths_rel() -> None:
    fpaths = ("var/foo/bar", "var/bar/foo")
    assert describe_files(fpaths) == "2 files"


def test_describe_files__iterable() -> None:
    fpaths = iter(("/var/foo/bar", "/var/foo/foo"))
    assert describe_files(fpaths) == "2 files in '/var/foo'"


def test_describe_files__non_str() -> None:
    with pytest.raises(TypeError):
        describe_files(1)


###############################################################################
###############################################################################
# Tests for 'describe_paired_files'


def test_describe_paired_files__single_file() -> None:
    fpath = "/var/foo/bar"
    assert describe_paired_files((fpath,), ()) == repr(fpath)


def test_describe_paired_files__identical_files() -> None:
    fpath = "/var/foo/bar"
    ftuple = (fpath,)
    assert describe_paired_files(ftuple, ftuple) == repr(fpath)


def test_describe_paired_files__same_path__similar_files() -> None:
    files_1 = ("foo/1_abc", "foo/1_def")
    files_2 = ("foo/1_ghi", "foo/1_jkl")
    expected = "'foo/1_???'"
    result = describe_paired_files(files_1, files_2)
    assert result == expected


def test_describe_paired_files__same_path__similar_files__different_prefixes() -> None:
    files_1 = ("foo/1_abc", "foo/1_def")
    files_2 = ("foo/2_ghi", "foo/2_jkl")
    expected = "'foo/[12]_???'"
    result = describe_paired_files(files_1, files_2)
    assert result == expected


def test_describe_paired_files__same_path__similar_files__too_different() -> None:
    files_1 = ("foo/1a_abc", "foo/1a_def")
    files_2 = ("foo/2b_ghi", "foo/2b_jkl")
    expected = "2 pair(s) of files in 'foo'"
    result = describe_paired_files(files_1, files_2)
    assert result == expected


def test_describe_paired_files__same_path__different_files() -> None:
    files_1 = ("foo/1_abc", "foo/2_def")
    files_2 = ("foo/3_ghi", "foo/4_jkl")
    expected = "2 pair(s) of files in 'foo'"
    result = describe_paired_files(files_1, files_2)
    assert result == expected


def test_describe_paired_files__same_path__different_file_lens() -> None:
    files_1 = ("foo/1_a", "foo/2_de")
    files_2 = ("foo/3_g", "foo/4_jk")
    expected = "2 pair(s) of files in 'foo'"
    result = describe_paired_files(files_1, files_2)
    assert result == expected


def test_describe_paired_files__different_path_and_files() -> None:
    files_1 = ("foo/1_abc", "bar/2_def")
    files_2 = ("zed/3_ghi", "not/4_jkl")
    expected = "2 pair(s) of files"
    result = describe_paired_files(files_1, files_2)
    assert result == expected


def test_describe_paired_files__files_1_longer() -> None:
    with pytest.raises(ValueError):
        describe_paired_files(("a", "b"), ("c",))


def test_describe_paired_files__files_2_longer() -> None:
    with pytest.raises(ValueError):
        describe_paired_files(("a",), ("b", "c"))


def test_describe_paired_files__non_str() -> None:
    with pytest.raises(TypeError):
        describe_paired_files((), 1)

    with pytest.raises(TypeError):
        describe_paired_files(1, ())

    with pytest.raises(TypeError):
        describe_paired_files(1, 1)



import nox

nox.options.default_venv_backend = "uv"
nox.options.sessions = [
    "style",
    "lints",
    "tests",
]


SOURCES = (
    "noxfile.py",
    "paleomix",
    "tests",
    "misc",
)


@nox.session
def style(session: nox.Session) -> None:
    session.install("--group", "lint", ".")
    # Replaces `black --check`
    session.run("ruff", "format", "--check", *SOURCES)
    # Replaces `isort --check-only`
    session.run("ruff", "check", "--select", "I", *SOURCES)


@nox.session
def lints(session: nox.Session) -> None:
    session.install("--group", "lint", ".")
    session.run("ruff", "check", *SOURCES)


@nox.session()
def tests(session: nox.Session) -> None:
    # Install in development mode to for coverage analysis
    session.install("--group", "dev", "-e", ".")
    session.run(
        "python3",
        # Run tests in development mode (enables extra checks)
        "-X",
        "dev",
        "-m",
        "pytest",
        "tests",
        "--cov",
        "paleomix",
        "--cov",
        "tests",
        "--cov-report=xml",
        "--cov-report=term-missing",
        "--no-cov-on-fail",
        # Exclude slow tests, including tests that run the pipeline via subprocess
        "-m",
        "not slow",
        "--quiet",
        # Treat warnings (deprections, etc.) as errors
        "-Werror",
        # Re-run failed tests, or all tests if there were no failures
        "--last-failed",
        "--last-failed-no-failures",
        "all",
    )


@nox.session(python=["3.8", "3.9", "3.10", "3.11", "3.12", "3.13"])
def full_tests(session: nox.Session) -> None:
    session.install("--group", "dev", ".")
    session.run(
        "python3",
        # Run tests in development mode (enables extra checks)
        "-X",
        "dev",
        "-m",
        "pytest",
        "tests",
        "--quiet",
        # Treat warnings (deprections, etc.) as errors
        "-Werror",
    )



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# Copyright (c) 2014 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import collections
import itertools
import os
import random
import sys
import tarfile
from io import TextIOWrapper

import pysam

import paleomix.common.argparse as argparse
import paleomix.common.bamfiles as bamtools
import paleomix.common.fileutils as fileutils
import paleomix.pipelines.zonkey.database as database
from paleomix.common.sampling import reservoir_sampling
from paleomix.common.sequences import NT_CODES

_TRANSITIONS = frozenset((("C", "T"), ("T", "C"), ("G", "A"), ("A", "G")))


def _filter_records(handle, flags=bamtools.EXCLUDED_FLAGS):
    for record in handle:
        if not record.flag & flags:
            yield record


class DownsampledBAM:
    def __init__(self, handle, downsample, included_references):
        self._records = collections.defaultdict(list)

        references = handle.references
        if len(references) != len(included_references):
            raise ValueError(
                "Length of 'included_references' must match the "
                "number of references in BAM file."
            )

        records = _filter_records(handle)
        for record in reservoir_sampling(records, downsample):
            key = references[record.tid]
            self._records[key].append(record)

        self.references = references

        self._records = dict(self._records)
        for value in self._records.values():
            value.sort(key=lambda rec: rec.pos)

    def fetch(self, chrom):
        return self._records.get(chrom, ())


class GenotypeSites:
    def __init__(self, records):
        last_chrom = None
        sites = []

        # chrom, pos, ref, ..., nucleotides
        for chrom, pos, line in records:
            # Convert pos from 1-based to 0-based (same as BAM.pos)
            sites.append((int(pos) - 1, line, []))

            assert last_chrom is None or chrom == last_chrom, (chrom, last_chrom)
            last_chrom = chrom

        sites.sort()
        self._sites = collections.deque(sites)

    def process(self, records, statistics):
        count_used = 0
        count_total = 0
        sites = self._sites
        for record_id, record in enumerate(records):
            count_total += 1

            # TODO: Check sorted
            while sites and sites[0][0] < record.pos:
                yield sites.popleft()

            sequence = record.seq
            sites_iter = iter(sites).__next__
            alignment_iter = iter(record.get_aligned_pairs()).__next__
            alignment_end = record.aend

            try:
                read_used = False
                site_pos, _, nucleotides = sites_iter()
                query_pos, ref_pos = alignment_iter()

                while alignment_end > site_pos:
                    if ref_pos is None or query_pos is None or site_pos > ref_pos:
                        query_pos, ref_pos = alignment_iter()
                    elif site_pos < ref_pos:
                        site_pos, _, nucleotides = sites_iter()
                    else:
                        assert ref_pos == site_pos, (ref_pos, site_pos)
                        nucleotide = sequence[query_pos]
                        if nucleotide != "N":
                            nucleotides.append((record_id, nucleotide))
                            read_used = True

                        query_pos, ref_pos = alignment_iter()
                        site_pos, _, nucleotides = sites_iter()
            except StopIteration:
                if not sites:
                    break
            finally:
                if read_used:
                    count_used += 1

        while sites:
            yield sites.popleft()

        statistics["n_reads"] += count_total
        statistics["n_reads_used"] += count_used


class GenotypeReader:
    def __init__(self, filename):
        self._tar_handle = tarfile.open(filename)
        self._handle = TextIOWrapper(self._tar_handle.extractfile("genotypes.txt"))
        self._header = self._handle.readline().rstrip("\r\n").split("\t")
        self.samples = self._header[-1].split(";")

    def __iter__(self):
        for chrom, records in itertools.groupby(
            self._read_records(), lambda rec: rec[0]
        ):
            sys.stderr.write("Reading contig %r information\n" % (chrom,))
            yield chrom, GenotypeSites(records)

    def _read_records(self):
        for line in self._handle:
            yield line.rstrip().split("\t", 2)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._handle.close()
        self._tar_handle.close()


def process_record(
    chrom,
    pos,
    line,
    nucleotides,
    statistics,
    records,
    out_incl_ts=sys.stdout,
    out_excl_ts=sys.stdout,
):
    # Filter reads that have already been used
    nucleotides = [
        (rec_id, nuc) for rec_id, nuc in nucleotides if rec_id not in records
    ]

    if not nucleotides:
        return
    elif len(nucleotides) == 1:
        # Avoid unnecessary random() call in 'random.choice'
        record_id, nucleotide = nucleotides[0]
    else:
        record_id, nucleotide = random.choice(nucleotides)

    # Fields are expected to contain at least 2 columns, the first being the
    # reference nucleotide, and the last being the sample genotypes
    _, encoded_genotypes = line.strip().rsplit("\t", 1)

    genotypes = []
    for encoded_nucleotide in encoded_genotypes:
        decoded_nucleotides = NT_CODES.get(encoded_nucleotide, ())
        if len(decoded_nucleotides) == 1:
            genotypes.append(decoded_nucleotides)
            genotypes.append(decoded_nucleotides)
        elif len(decoded_nucleotides) == 2:
            genotypes.extend(decoded_nucleotides)
        else:
            raise ValueError(
                "Invalid nucleotide, not bi-allelic: %r" % (encoded_nucleotide,)
            )

    if nucleotide not in genotypes:
        # Exclude SNPs not observed in the reference panel
        return

    # Convert from 0-based to 1-based
    pos += 1

    # Chromosome, SNP identifier, (dummy) pos in (centi)Morgans, position
    output = [chrom, "chr{}_{}".format(chrom, pos), "0", str(pos)]
    output.extend(genotypes)
    output.append(nucleotide)
    output.append(nucleotide)
    output = " ".join(output)

    statistics["n_sites_incl_ts"] += 1
    out_incl_ts.write("{}\n".format(output))

    if tuple(set(genotypes)) not in _TRANSITIONS:
        statistics["n_sites_excl_ts"] += 1
        out_excl_ts.write("{}\n".format(output))

    records.add(record_id)


def write_tfam(filename, data, samples, bam_sample):
    with open(filename, "w") as handle:
        for key in samples:
            row = data.samples[key]
            sex = {"MALE": 1, "FEMALE": 2, "NA": 0}[row["Sex"].upper()]
            # Family, Individual, Paternal ID, Maternal ID, Sex, Phenotype
            handle.write("{0} {0} 0 0 {1} -9\n".format(key, sex))

        handle.write("{0} {0} 0 0 0 -9\n".format(bam_sample))


def write_summary(args, filename, statistics):
    with open(filename, "w") as handle:
        handle.write("name: %s\n" % (args.name,))
        handle.write("filename: %s\n" % (os.path.abspath(args.bam),))

        for key in ("n_reads", "n_reads_used", "n_sites_incl_ts", "n_sites_excl_ts"):
            handle.write("%s: %s\n" % (key, statistics.get(key, "MISSING")))


def process_bam(args, data, bam_handle, mapping):
    reverse_mapping = dict(zip(mapping.values(), mapping))
    raw_references = bam_handle.references
    references = [reverse_mapping.get(name, name) for name in raw_references]

    if args.downsample:
        sys.stderr.write("Downsampling to at most %i BAM records\n" % (args.downsample))
        bam_handle = DownsampledBAM(bam_handle, args.downsample, references)

    statistics = {
        "n_reads": 0,
        "n_reads_used": 0,
        "n_sites_incl_ts": 0,
        "n_sites_excl_ts": 0,
    }

    fileutils.make_dirs(args.root)

    with open(os.path.join(args.root, "incl_ts.tped"), "w") as output_incl:
        with open(os.path.join(args.root, "excl_ts.tped"), "w") as output_excl:
            with GenotypeReader(args.database) as reader:
                for ref, sites in reader:
                    records = set()
                    raw_ref = raw_references[references.index(ref)]

                    sys.stderr.write("Reading %r from BAM\n" % (raw_ref,))
                    raw_sites = bam_handle.fetch(raw_ref)
                    for pos, line, nucleotides in sites.process(raw_sites, statistics):
                        process_record(
                            ref,
                            pos,
                            line,
                            nucleotides,
                            out_incl_ts=output_incl,
                            out_excl_ts=output_excl,
                            statistics=statistics,
                            records=records,
                        )

                write_summary(
                    args,
                    os.path.join(args.root, "common.summary"),
                    statistics=statistics,
                )
                write_tfam(
                    os.path.join(args.root, "common.tfam"),
                    data,
                    reader.samples,
                    args.name,
                )


def parse_args(argv):
    parser = argparse.ArgumentParser(prog="paleomix zonkey:tped")
    parser.add_argument(
        "root",
        metavar="output_folder",
        help="Output folder in which output files are "
        "to be placed; is created if it does not "
        "already exist.",
    )
    parser.add_argument("database", help="Zonkey database file.")
    parser.add_argument("bam", help="Sorted BAM file.")
    parser.add_argument(
        "--seed",
        type=int,
        help="RNG seed used when downsampling reads; "
        "defaults to using system time as seed.",
    )
    parser.add_argument(
        "--downsample",
        type=int,
        default=0,
        help="Sample N reads from the input BAM file, before building the TPED file. "
        "If not set, or set to zero, all reads are used",
    )
    parser.add_argument(
        "--name", default="Sample", help="Name of sample to be used in output."
    )

    return parser.parse_args(argv)


def main(argv):
    args = parse_args(argv)
    random.seed(args.seed)

    print("Reading reference information from %r" % (args.database,))

    try:
        data = database.ZonkeyDB(args.database)
    except database.ZonkeyDBError as error:
        sys.stderr.write(
            "Error reading database file %r:\n%s\n" % (args.database, error)
        )
        return 1

    with pysam.AlignmentFile(args.bam) as bam_handle:
        bam_info = data.validate_bam_handle(bam_handle)
        if not bam_info:
            return 1
        elif not bam_info.is_nuclear:
            sys.stderr.write(
                "ERROR: BAM file does not contain identifiable nuclear alignments.\n"
            )
            return 1

        process_bam(args, data, bam_handle, bam_info.nuclear_contigs)

    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import itertools

import pytest

from paleomix.common.sequences import (
    complement,
    encode_genotype,
    reverse_complement,
    split,
)

###############################################################################
###############################################################################
# Tests for 'complement'

_REF_SRC = "ACGTMRWSYKVHDBNX"
_REF_DST = "TGCAKYWSRMBDHVNX"


@pytest.mark.parametrize("src, dst", zip(_REF_SRC, _REF_DST))
def test_complement__single_nt(src, dst):
    assert complement(src) == dst
    assert complement(src.lower()) == dst.lower()


def test_complement__multiple_nts_upper():
    assert complement(_REF_SRC) == _REF_DST


def test_complement__multiple_nts_lower():
    assert complement(_REF_SRC.lower()) == _REF_DST.lower()


def test_complement__multiple_nts_mixed_case():
    assert complement("aGtCn") == "tCaGn"


###############################################################################
###############################################################################
# Tests for 'complement'


def test_reverse_complement():
    assert reverse_complement(_REF_SRC) == _REF_DST[::-1]


###############################################################################
###############################################################################
# Tests for 'encode_genotype'

_IUB_SRC = (
    "A",
    "C",
    "G",
    "T",
    "AC",
    "AG",
    "AT",
    "CG",
    "CT",
    "GT",
    "ACG",
    "ACT",
    "AGT",
    "CGT",
    "ACGT",
)
_IUB_DST = "ACGTMRWSYKVHDB"


@pytest.mark.parametrize("src, dst", zip(_IUB_SRC, _IUB_DST))
def test_genotype__permutations(src, dst):
    for seq in itertools.permutations(src):
        assert encode_genotype("".join(src)) == dst


@pytest.mark.parametrize("value", ("a", "At", "Z", "+"))
def test_genotype__bad_input(value):
    with pytest.raises(ValueError):
        encode_genotype(value)


@pytest.mark.parametrize("sequence", ("CT", "C,T", ",C,T", "C,T,", ",C,T,"))
def test_comma_or_not(sequence):
    assert encode_genotype(sequence) == "Y"


###############################################################################
###############################################################################
# Tests for 'split'


def test_split__empty_sequence():
    assert split("") == {"1": "", "2": "", "3": ""}


def test_split__no_split_by():
    with pytest.raises(ValueError, match="No split_by specified"):
        split("", split_by="")


def test_split__single_group():
    assert split("ACGCAT", "111") == {"1": "ACGCAT"}


def test_split__two_groups():
    assert split("ACGCAT", "112") == {"1": "ACCA", "2": "GT"}


def test_split__three_groups():
    expected = {"1": "AC", "2": "CA", "3": "GT"}
    assert split("ACGCAT", "123") == expected
    assert split("ACGCAT") == expected


def test_split__empty_group():
    expected = {"1": "A", "2": "C", "3": ""}
    assert split("AC") == expected


def test_split__partial_group():
    expected = {"1": "AA", "2": "CA", "3": "G"}
    assert split("ACGAA") == expected



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import collections
import errno
import logging
import os

import paleomix.common.versions as versions
from paleomix.common.fileutils import missing_executables, missing_files
from paleomix.common.utilities import safe_coerce_to_frozenset


class FileStatusCache:
    """Cache used to avoid repeatedly checking the state (existence / mtime) of
    files required / generated by nodes. A new cache is generated for every
    operation (e.g. refreshing all states / manually setting the state of a
    node) to avoid relying on the filesystem staying consistent for long
    periods of time.
    """

    def __init__(self):
        self._stat_cache = {}

    def files_exist(self, fpaths):
        """Returns true if all paths listed in fpaths exist."""
        return all((self._get_state(fpath) is not None) for fpath in fpaths)

    def missing_files(self, fpaths):
        """Returns a list of paths in fpaths that do not exist."""
        return [fpath for fpath in fpaths if (self._get_state(fpath) is None)]

    def are_files_outdated(self, input_files, output_files):
        """Returns true if any 'input' files have a time-stamp that post-date
        any time-stamp for the 'output' files, indicating that one or more of
        the 'input' files have changed since the creation of the 'output'.

        The function also returns true if any files are missing, as this would
        indicate either the 'output' files or both the 'input' and 'output'
        files would need to be rebuilt.
        """
        input_timestamps = []
        if not self._get_states(input_files, input_timestamps):
            return True

        output_timestamps = []
        if not self._get_states(output_files, output_timestamps):
            return True

        return max(input_timestamps) > min(output_timestamps)

    def _get_states(self, filenames, dst):
        """Collects the mtimes for a set of filenames, returning true if all
        could be collected, and aborting early and returning false otherwise.
        """
        for filename in filenames:
            timestamp = self._get_state(filename)
            if timestamp is None:
                return False

            dst.append(timestamp)

        return True

    def _get_state(self, fpath):
        """Returns the mtime of a path, or None if the path does not exist."""
        if fpath not in self._stat_cache:
            try:
                mtime = os.path.getmtime(fpath)
            except OSError as error:
                if error.errno != errno.ENOENT:
                    raise
                mtime = None
            self._stat_cache[fpath] = mtime
        return self._stat_cache[fpath]


class NodeGraphError(RuntimeError):
    pass


class NodeGraph:
    NUMBER_OF_STATES = 6
    DONE, RUNNING, RUNABLE, QUEUED, OUTDATED, ERROR = range(NUMBER_OF_STATES)

    def __init__(
        self,
        nodes,
        allow_missing_files=False,
        cache_factory=FileStatusCache,
    ):
        self._cache_factory = cache_factory
        self._states = {}
        self._state_counts = [0] * self.NUMBER_OF_STATES

        nodes = safe_coerce_to_frozenset(nodes)

        self._logger = logging.getLogger(__name__)
        self._reverse_dependencies = collections.defaultdict(set)
        self._collect_reverse_dependencies(nodes, self._reverse_dependencies, set())
        self._intersections = {}
        self._top_nodes = [
            node
            for (node, rev_deps) in self._reverse_dependencies.items()
            if not rev_deps
        ]

        self._logger.info("Checking file dependencies")
        if not self._check_file_dependencies(self._reverse_dependencies):
            raise NodeGraphError("Aborting due to input/output file error")

        self._logger.info("Checking for auxiliary files")
        if not self._check_auxiliary_files(self._reverse_dependencies):
            if not allow_missing_files:
                raise NodeGraphError(
                    "Please refer to the PALEOMIX installation instructions at "
                    "https://paleomix.readthedocs.io/en/stable/"
                )

        self._logger.info("Checking required software")
        if not self._check_version_requirements(self._reverse_dependencies):
            if not allow_missing_files:
                raise NodeGraphError(
                    "Please refer to the PALEOMIX installation instructions at "
                    "https://paleomix.readthedocs.io/en/stable/"
                )

        self._logger.info("Determining states")
        self._refresh_states()
        self._logger.info("Ready")

    def get_node_state(self, node):
        return self._states[node]

    def set_node_state(self, node, state):
        if state not in (NodeGraph.RUNNING, NodeGraph.ERROR, NodeGraph.DONE):
            raise ValueError("Invalid state: %r" % (state,))
        old_state = self._states[node]
        if state == old_state:
            return

        self._states[node] = state
        self._state_counts[old_state] -= 1
        self._state_counts[state] += 1

        self._log_node_changes(node, old_state, state)

        intersections = self._calculate_intersections(node)

        # Not all nodes may need to be updated, but we still need to
        # traverse the "graph" (using the intersection counts) in order
        # to ensure that all nodes that need to be updated are updated.
        requires_update = dict.fromkeys(intersections, False)
        for dependency in self._reverse_dependencies[node]:
            requires_update[dependency] = True

        cache = self._cache_factory()
        while any(requires_update.values()):
            for node, count in tuple(intersections.items()):
                if not count:
                    has_changed = False
                    if requires_update[node]:
                        old_state = self._states.pop(node)
                        new_state = self._update_node_state(node, cache)
                        has_changed |= new_state != old_state

                        self._state_counts[old_state] -= 1
                        self._state_counts[new_state] += 1

                    for dependency in self._reverse_dependencies[node]:
                        intersections[dependency] -= 1
                        requires_update[dependency] |= has_changed

                    intersections.pop(node)
                    requires_update.pop(node)

    def __iter__(self):
        """Returns a graph of nodes."""
        return iter(self._top_nodes)

    def iterflat(self):
        return iter(self._reverse_dependencies)

    def get_state_counts(self):
        return list(self._state_counts)

    def _refresh_states(self):
        states = {}
        cache = self._cache_factory()
        for node, state in self._states.items():
            if state in (self.ERROR, self.RUNNING):
                states[node] = state
        self._states = states

        for node in self._reverse_dependencies:
            self._update_node_state(node, cache)

        state_counts = [0] * self.NUMBER_OF_STATES
        for state in states.values():
            state_counts[state] += 1
        self._state_counts = state_counts

    def _log_node_changes(self, node, old_state, new_state):
        if new_state in (self.RUNNING, self.DONE):
            running = self._state_counts[self.RUNNING]
            remaining = (
                sum(self._state_counts)
                - self._state_counts[self.DONE]
                - self._state_counts[self.ERROR]
            )

            if new_state == self.RUNNING:
                self._logger.info("[%i/%i] Started %s", running, remaining, node)
            elif new_state == self.DONE:
                self._logger.info("[%i/%i] Finished %s", running, remaining, node)

    def _calculate_intersections(self, for_node):
        def count_nodes(node, counts):
            for node in self._reverse_dependencies[node]:
                if node in counts:
                    counts[node] += 1
                else:
                    counts[node] = 1
                    count_nodes(node, counts)
            return counts

        if for_node not in self._intersections:
            counts = count_nodes(for_node, {})
            for dependency in self._reverse_dependencies[for_node]:
                counts[dependency] -= 1
            self._intersections[for_node] = counts

        return dict(self._intersections[for_node])

    def _update_node_state(self, node, cache):
        if node in self._states:
            return self._states[node]

        # Update sub-nodes before checking for fixed states
        dependency_states = set((NodeGraph.DONE,))
        for dependency in node.dependencies:
            dependency_states.add(self._update_node_state(dependency, cache))

        state = max(dependency_states)
        if state == NodeGraph.DONE:
            if not self.is_done(node, cache):
                state = NodeGraph.RUNABLE
            elif not cache.files_exist(node.input_files):
                # Somehow the input files have gone missing, despite the
                # dependent nodes being done; this implies this node is
                # outdated, since the input-files should be re-generated, but
                # obviously it is not possible to run it at this point.
                missing = cache.missing_files(node.input_files)
                self._logger.error(
                    "Input file(s) missing for node; may have been moved while the "
                    "pipeline was running. Cannot proceed:\n"
                    "    Node = %s\n    Files = %s\n",
                    node,
                    "\n            ".join(missing),
                )
                state = NodeGraph.ERROR
            elif self.is_outdated(node, cache):
                state = NodeGraph.RUNABLE
        elif state in (NodeGraph.RUNNING, NodeGraph.RUNABLE, NodeGraph.QUEUED):
            if self.is_done(node, cache):
                state = NodeGraph.OUTDATED
            else:
                state = NodeGraph.QUEUED
        self._states[node] = state

        return state

    @classmethod
    def is_done(cls, node, cache):
        """Returns true if the node itself is done; this only implies that the
        output files generated by this node exists. The files themselves may
        be outdated.
        """
        return cache.files_exist(node.output_files)

    @classmethod
    def is_outdated(cls, node, cache):
        """Returns true if the not is not done or if one or more of the input
        files appear to have been changed since the creation of the output
        files (based on the timestamps). A node that lacks either input or
        output files is never considered outdated.
        """
        if not (node.input_files and node.output_files):
            return False

        return cache.are_files_outdated(node.input_files, node.output_files)

    def _check_required_executables(self, nodes):
        exec_filenames = set()
        for node in nodes:
            exec_filenames.update(node.executables)

            # Requirements may include executables not invoked directly
            for requirement in node.requirements:
                if isinstance(requirement, versions.RequirementObj):
                    executable = requirement.executable
                    if executable is not None:
                        exec_filenames.add(executable)

        missing_exec = missing_executables(exec_filenames)
        if missing_exec:
            self._logger.error("Required executable(s) not found:")
            for name in sorted(missing_exec):
                self._logger.error(" - %s", name)

        return not missing_exec

    def _check_version_requirements(self, nodes):
        if not self._check_required_executables(nodes):
            return False

        exec_requirements = set()
        for node in nodes:
            exec_requirements.update(node.requirements)

        def _key_func(reqobj):
            # Sort priority in decreasing order, name in increasing order
            return (-reqobj.priority, reqobj.name)

        any_errors = False
        for requirement in sorted(exec_requirements, key=_key_func):
            try:
                name = requirement.name
                version = ".".join(str(value) for value in requirement.version)
                if version:
                    name = "%s v%s" % (name, version)

                self._logger.info(" - Found %s", name)

                requirement()
            except versions.VersionRequirementError as error:
                any_errors = True
                self._logger.error(error)
            except OSError as error:
                any_errors = True
                self._logger.error(
                    "Could not check version for %s:\n\t%s" % (requirement.name, error)
                )

        return not any_errors

    def _check_file_dependencies(self, nodes, max_errors=10):
        input_files = collections.defaultdict(set)
        output_files = collections.defaultdict(set)
        for node in nodes:
            for filename in node.input_files:
                input_files[filename].add(node)

            for filename in node.output_files:
                output_files[filename].add(node)

        any_errors = False
        if not self._check_output_files(output_files, max_errors):
            any_errors = True

        if not self._check_input_files(input_files, output_files, nodes, max_errors):
            any_errors = True

        return not any_errors

    def _check_output_files(self, output_files, max_errors=10):
        """Checks dict of output files to nodes for cases where
        multiple nodes create the same output file.

        The directory component of paths are realized in order to
        detect cases where nodes create the same file, but via
        different paths (e.g. due to relative/absolute paths, or
        due to use of symbolic links). Since output files are
        replaced, not modified in place, it is not necessary to
        compare files themselves."""
        dirpath_cache = {}
        filepaths = collections.defaultdict(list)
        for filename, nodes in output_files.items():
            dirpath, filename = os.path.split(filename)
            if dirpath not in dirpath_cache:
                dirpath_cache[dirpath] = os.path.realpath(dirpath)

            filepath = os.path.join(dirpath_cache[dirpath], filename)
            filepaths[filepath].extend(nodes)

        clobbered_files = {
            key: values for key, values in filepaths.items() if len(values) > 1
        }

        if clobbered_files:
            for filename, nodes in sorted(clobbered_files.items()):
                self._logger.error("Multiple nodes write to file %r:", filename)
                self._logger.debug("depended on by ")
                for line in _summarize_nodes(nodes):
                    self._logger.debug("  %s", line)

                self._logger.error(
                    "This is probably a bug in PALEOMIX; please report at "
                    "https://github.com/MikkelSchubert/paleomix/issues/new"
                )

        return not clobbered_files

    def _check_input_files(self, input_files, output_files, nodes, max_errors=10):
        dependencies = self._collect_dependencies(nodes, {})
        any_errors = False

        for filename, nodes in sorted(input_files.items(), key=lambda v: v[0]):
            if filename in output_files:
                producers = output_files[filename]
                bad_nodes = set()
                for consumer in nodes:
                    if not (producers & dependencies[consumer]):
                        bad_nodes.add(consumer)

                if bad_nodes:
                    any_errors = True
                    producer = next(iter(producers))
                    bad_nodes = _summarize_nodes(bad_nodes)
                    self._logger.error(
                        "Tasks depends on file, but not on the task creating it: %r",
                        filename,
                    )

                    self._logger.debug("  produced by %s", producer)
                    self._logger.debug("  depended on by")
                    for line in _summarize_nodes(bad_nodes):
                        self._logger.debug("    %s", line)

                    self._logger.error(
                        "This is probably a bug in PALEOMIX; please report at "
                        "https://github.com/MikkelSchubert/paleomix/issues/new"
                    )

            elif not os.path.exists(filename):
                any_errors = True
                self._logger.error("Required input file does not exist: %r", filename)
                for line in _summarize_nodes(nodes):
                    self._logger.debug("  required by %s", line)

        return not any_errors

    def _check_auxiliary_files(self, nodes):
        auxiliary_files = set()
        for node in nodes:
            auxiliary_files.update(node.auxiliary_files)

        missing_aux_files = missing_files(auxiliary_files)
        if missing_aux_files:
            self._logger.error("Required files not found:")
            for name in sorted(missing_aux_files):
                self._logger.error(" - %s", name)

        return not missing_aux_files

    @classmethod
    def _collect_dependencies(cls, nodes, dependencies):
        for node in nodes:
            if node not in dependencies:
                subnodes = node.dependencies
                if not subnodes:
                    dependencies[node] = frozenset()
                    continue

                cls._collect_dependencies(subnodes, dependencies)

                collected = set(subnodes)
                for subnode in subnodes:
                    collected.update(dependencies[subnode])
                dependencies[node] = frozenset(collected)

        return dependencies

    @classmethod
    def _collect_reverse_dependencies(cls, lst, rev_dependencies, processed):
        for node in lst:
            if node not in processed:
                processed.add(node)

                # Initialize default-dict
                rev_dependencies[node]

                subnodes = node.dependencies
                for dependency in subnodes:
                    rev_dependencies[dependency].add(node)
                cls._collect_reverse_dependencies(subnodes, rev_dependencies, processed)


def _summarize_nodes(nodes):
    nodes = list(sorted(set(map(str, nodes))))
    if len(nodes) > 4:
        nodes = nodes[:5] + ["and %i more nodes" % len(nodes)]
    return nodes
